from unittest.mock import MagicMock

import pytest
from pymetastore.hive_metastore.ttypes import Date, Decimal
from pymetastore.htypes import (
    HCharType,
    HDecimalType,
    HListType,
    HMapType,
    HPrimitiveType,
    HStructType,
    HUnionType,
    HVarcharType,
    PrimitiveCategory,
)
from pymetastore.metastore import HMS, HColumn
from pymetastore.stats import (
    BinaryTypeStats,
    BooleanTypeStats,
    DateTypeStats,
    DecimalTypeStats,
    DoubleTypeStats,
    LongTypeStats,
    StringTypeStats,
)

from recap.clients.hive_metastore import HiveMetastoreClient
from recap.types import (
    BoolType,
    BytesType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    StringType,
    StructType,
    UnionType,
)


def test_struct():
    mock_client = MagicMock(spec=HMS)

    class MockTable:
        name = "dummy_table"
        columns = [
            HColumn("col1", HPrimitiveType(PrimitiveCategory.BOOLEAN)),
            HColumn("col2", HPrimitiveType(PrimitiveCategory.BYTE)),
            HColumn("col3", HPrimitiveType(PrimitiveCategory.SHORT)),
            HColumn("col4", HPrimitiveType(PrimitiveCategory.INT)),
            HColumn("col5", HPrimitiveType(PrimitiveCategory.LONG)),
            HColumn("col6", HPrimitiveType(PrimitiveCategory.FLOAT)),
            HColumn("col7", HPrimitiveType(PrimitiveCategory.DOUBLE)),
            HColumn("col8", HPrimitiveType(PrimitiveCategory.STRING)),
            HColumn("col9", HPrimitiveType(PrimitiveCategory.BINARY)),
            HColumn("col10", HDecimalType(10, 2)),
            HColumn("col11", HVarcharType(100)),
            HColumn("col12", HCharType(100)),
            HColumn("col13", HPrimitiveType(PrimitiveCategory.VOID)),
        ]

    mock_client.get_table.return_value = MockTable

    client = HiveMetastoreClient(mock_client)
    result = client.schema("dummy_database", "dummy_table")

    # Check that the schema was converted correctly.
    assert isinstance(result, StructType)
    assert len(result.fields) == 13

    # Validate each column in the order defined in MockTable
    assert isinstance(result.fields[0], UnionType)
    assert isinstance(result.fields[0].types[1], BoolType)
    assert result.fields[0].extra_attrs["name"] == "col1"

    assert isinstance(result.fields[1], UnionType)
    assert isinstance(result.fields[1].types[1], IntType)
    assert result.fields[1].types[1].bits == 8
    assert result.fields[1].extra_attrs["name"] == "col2"

    assert isinstance(result.fields[2], UnionType)
    assert isinstance(result.fields[2].types[1], IntType)
    assert result.fields[2].types[1].bits == 16
    assert result.fields[2].extra_attrs["name"] == "col3"

    assert isinstance(result.fields[3], UnionType)
    assert isinstance(result.fields[3].types[1], IntType)
    assert result.fields[3].types[1].bits == 32
    assert result.fields[3].extra_attrs["name"] == "col4"

    assert isinstance(result.fields[4], UnionType)
    assert isinstance(result.fields[4].types[1], IntType)
    assert result.fields[4].types[1].bits == 64
    assert result.fields[4].extra_attrs["name"] == "col5"

    assert isinstance(result.fields[5], UnionType)
    assert isinstance(result.fields[5].types[1], FloatType)
    assert result.fields[5].types[1].bits == 32
    assert result.fields[5].extra_attrs["name"] == "col6"

    assert isinstance(result.fields[6], UnionType)
    assert isinstance(result.fields[6].types[1], FloatType)
    assert result.fields[6].types[1].bits == 64
    assert result.fields[6].extra_attrs["name"] == "col7"

    assert isinstance(result.fields[7], UnionType)
    assert isinstance(result.fields[7].types[1], StringType)
    assert result.fields[7].types[1].bytes_ == 9_223_372_036_854_775_807
    assert result.fields[7].extra_attrs["name"] == "col8"

    assert isinstance(result.fields[8], UnionType)
    assert isinstance(result.fields[8].types[1], BytesType)
    assert result.fields[8].types[1].bytes_ == 2_147_483_647
    assert result.fields[8].extra_attrs["name"] == "col9"

    assert isinstance(result.fields[9], UnionType)
    assert isinstance(result.fields[9].types[1], BytesType)
    assert result.fields[9].types[1].bytes_ == 16
    assert result.fields[9].types[1].extra_attrs["precision"] == 10
    assert result.fields[9].types[1].extra_attrs["scale"] == 2
    assert result.fields[9].extra_attrs["name"] == "col10"

    assert isinstance(result.fields[10], UnionType)
    assert isinstance(result.fields[10].types[1], StringType)
    assert result.fields[10].types[1].bytes_ == 100
    assert result.fields[10].extra_attrs["name"] == "col11"

    assert isinstance(result.fields[11], UnionType)
    assert isinstance(result.fields[11].types[1], StringType)
    assert result.fields[11].types[1].bytes_ == 100
    assert not result.fields[11].types[1].variable
    assert result.fields[11].extra_attrs["name"] == "col12"

    assert isinstance(result.fields[12], NullType)
    assert result.fields[12].extra_attrs["name"] == "col13"


def test_struct_with_struct_type():
    mock_client = MagicMock(spec=HMS)

    class MockTable:
        name = "dummy_table"
        columns = [
            HColumn(
                "col1",
                HStructType(
                    names=["sub_col1", "sub_col2"],
                    types=[
                        HPrimitiveType(PrimitiveCategory.BOOLEAN),
                        HPrimitiveType(PrimitiveCategory.INT),
                    ],
                ),
            ),
        ]

    mock_client.get_table.return_value = MockTable

    client = HiveMetastoreClient(mock_client)
    result = client.schema("dummy_database", "dummy_table")

    # Check that the schema was converted correctly.
    assert isinstance(result, StructType)
    assert len(result.fields) == 1

    # Validate the column with StructType
    assert isinstance(result.fields[0], UnionType)
    assert isinstance(result.fields[0].types[1], StructType)
    assert result.fields[0].extra_attrs["name"] == "col1"

    # Validate sub-columns of the struct
    struct_field = result.fields[0].types[1]
    assert len(struct_field.fields) == 2

    # Validate sub_col1
    assert isinstance(struct_field.fields[0], UnionType)
    assert isinstance(struct_field.fields[0].types[1], BoolType)
    assert struct_field.fields[0].extra_attrs["name"] == "sub_col1"

    # Validate sub_col2
    assert isinstance(struct_field.fields[1], UnionType)
    assert isinstance(struct_field.fields[1].types[1], IntType)
    assert struct_field.fields[1].types[1].bits == 32
    assert struct_field.fields[1].extra_attrs["name"] == "sub_col2"


def test_struct_with_list_type():
    mock_client = MagicMock(spec=HMS)

    class MockTable:
        name = "dummy_table"
        columns = [
            HColumn("col1", HListType(HPrimitiveType(PrimitiveCategory.BOOLEAN))),
        ]

    mock_client.get_table.return_value = MockTable

    client = HiveMetastoreClient(mock_client)
    result = client.schema("dummy_database", "dummy_table")

    assert isinstance(result, StructType)
    assert len(result.fields) == 1
    assert isinstance(result.fields[0], UnionType)
    assert isinstance(result.fields[0].types[1], ListType)
    assert isinstance(result.fields[0].types[1].values, UnionType)
    assert isinstance(result.fields[0].types[1].values.types[1], BoolType)
    assert result.fields[0].extra_attrs["name"] == "col1"


def test_struct_with_map_type():
    mock_client = MagicMock(spec=HMS)

    class MockTable:
        name = "dummy_table"
        columns = [
            HColumn(
                "col1",
                HMapType(
                    HPrimitiveType(PrimitiveCategory.STRING),
                    HPrimitiveType(PrimitiveCategory.INT),
                ),
            ),
        ]

    mock_client.get_table.return_value = MockTable

    client = HiveMetastoreClient(mock_client)
    result = client.schema("dummy_database", "dummy_table")

    assert isinstance(result, StructType)
    assert len(result.fields) == 1
    assert isinstance(result.fields[0], UnionType)
    assert isinstance(result.fields[0].types[1], MapType)
    assert isinstance(result.fields[0].types[1].keys, UnionType)
    assert isinstance(result.fields[0].types[1].values, UnionType)
    assert isinstance(result.fields[0].types[1].keys.types[1], StringType)
    assert isinstance(result.fields[0].types[1].values.types[1], IntType)
    assert result.fields[0].extra_attrs["name"] == "col1"


def test_struct_with_nested_struct_type():
    mock_client = MagicMock(spec=HMS)

    class MockTable:
        name = "dummy_table"
        columns = [
            HColumn(
                "col1",
                HStructType(
                    names=["sub_col1", "sub_col2"],
                    types=[
                        HStructType(
                            names=["nested_col1", "nested_col2"],
                            types=[
                                HPrimitiveType(PrimitiveCategory.BOOLEAN),
                                HPrimitiveType(PrimitiveCategory.INT),
                            ],
                        ),
                        HPrimitiveType(PrimitiveCategory.STRING),
                    ],
                ),
            ),
        ]

    mock_client.get_table.return_value = MockTable

    client = HiveMetastoreClient(mock_client)
    result = client.schema("dummy_database", "dummy_table")

    # Check that the schema was converted correctly.
    assert isinstance(result, StructType)
    assert len(result.fields) == 1

    # Validate the column with StructType
    assert isinstance(result.fields[0], UnionType)
    assert isinstance(result.fields[0].types[1], StructType)
    assert result.fields[0].extra_attrs["name"] == "col1"

    # Validate sub-columns of the struct
    struct_field = result.fields[0].types[1]
    assert len(struct_field.fields) == 2

    # Validate sub_col1, which is another struct
    assert isinstance(struct_field.fields[0], UnionType)
    assert isinstance(struct_field.fields[0].types[1], StructType)
    assert struct_field.fields[0].extra_attrs["name"] == "sub_col1"

    # Validate nested_col1 and nested_col2 within sub_col1
    nested_struct_field = struct_field.fields[0].types[1]
    assert len(nested_struct_field.fields) == 2

    assert isinstance(nested_struct_field.fields[0], UnionType)
    assert isinstance(nested_struct_field.fields[0].types[1], BoolType)
    assert nested_struct_field.fields[0].extra_attrs["name"] == "nested_col1"

    assert isinstance(nested_struct_field.fields[1], UnionType)
    assert isinstance(nested_struct_field.fields[1].types[1], IntType)
    assert nested_struct_field.fields[1].types[1].bits == 32
    assert nested_struct_field.fields[1].extra_attrs["name"] == "nested_col2"

    # Validate sub_col2
    assert isinstance(struct_field.fields[1], UnionType)
    assert isinstance(struct_field.fields[1].types[1], StringType)
    assert struct_field.fields[1].types[1].bytes_ == 9_223_372_036_854_775_807
    assert struct_field.fields[1].extra_attrs["name"] == "sub_col2"


def test_struct_with_union_type():
    mock_client = MagicMock(spec=HMS)

    class MockTable:
        name = "dummy_table"
        columns = [
            HColumn(
                "col1",
                HUnionType(
                    [
                        HPrimitiveType(PrimitiveCategory.BOOLEAN),
                        HPrimitiveType(PrimitiveCategory.INT),
                    ]
                ),
            ),
        ]

    mock_client.get_table.return_value = MockTable

    client = HiveMetastoreClient(mock_client)
    result = client.schema("dummy_database", "dummy_table")

    assert isinstance(result, StructType)
    assert len(result.fields) == 1
    assert isinstance(result.fields[0], UnionType)
    assert result.fields[0].extra_attrs["name"] == "col1"
    assert len(result.fields[0].types) == 3  # null type + two types from HUnionType
    assert isinstance(result.fields[0].types[0], NullType)
    assert isinstance(result.fields[0].types[1], BoolType)
    assert isinstance(result.fields[0].types[2], IntType)


def test_get_table_stats():
    mock_client = MagicMock(spec=HMS)

    class MockTable:
        name = "dummy_table"
        columns = [
            HColumn("col1", HPrimitiveType(PrimitiveCategory.BOOLEAN)),
            HColumn("col2", HPrimitiveType(PrimitiveCategory.LONG)),
            HColumn("col3", HPrimitiveType(PrimitiveCategory.DOUBLE)),
            HColumn("col4", HPrimitiveType(PrimitiveCategory.STRING)),
            HColumn("col5", HPrimitiveType(PrimitiveCategory.BINARY)),
            HColumn("col6", HPrimitiveType(PrimitiveCategory.DATE)),
            HColumn("col7", HDecimalType(precision=10, scale=5)),
        ]

    class MockBooleanStats:
        columnName = "col1"
        stats = BooleanTypeStats(numFalses=2, numTrues=1, numNulls=3)

    class MockLongStats:
        columnName = "col2"
        stats = LongTypeStats(lowValue=0, highValue=100, numNulls=10, cardinality=90)

    class MockDoubleStats:
        columnName = "col3"
        stats = DoubleTypeStats(
            cardinality=85,
            lowValue=0.1,
            highValue=100.1,
            numNulls=15,
        )

    class MockStringStats:
        columnName = "col4"
        stats = StringTypeStats(
            avgColLen=10,
            maxColLen=100,
            cardinality=80,
            numNulls=20,
        )

    class MockBinaryStats:
        columnName = "col5"
        stats = BinaryTypeStats(avgColLen=500, maxColLen=1000, numNulls=30)

    class MockDateStats:
        columnName = "col6"
        stats = DateTypeStats(
            cardinality=90,
            lowValue=Date(123),
            highValue=Date(456),
            numNulls=40,
        )

    class MockDecimalStats:
        columnName = "col7"
        stats = DecimalTypeStats(
            cardinality=95,
            lowValue=Decimal("0.1"),
            highValue=Decimal("100.1"),
            numNulls=50,
        )

    mock_client.get_table.return_value = MockTable
    mock_client.get_table_stats.return_value = [
        MockBooleanStats,
        MockLongStats,
        MockDoubleStats,
        MockStringStats,
        MockBinaryStats,
        MockDateStats,
        MockDecimalStats,
    ]

    client = HiveMetastoreClient(mock_client)
    result = client.schema("dummy_database", "dummy_table", True)

    assert isinstance(result, StructType)
    assert len(result.fields) == 7

    # Boolean stats check
    assert result.fields[0].extra_attrs["name"] == "col1"
    assert result.fields[0].extra_attrs["false_count"] == 2
    assert result.fields[0].extra_attrs["true_count"] == 1
    assert result.fields[0].extra_attrs["null_count"] == 3

    # Long stats check
    assert result.fields[1].extra_attrs["name"] == "col2"
    assert result.fields[1].extra_attrs["low"] == 0
    assert result.fields[1].extra_attrs["high"] == 100
    assert result.fields[1].extra_attrs["null_count"] == 10
    assert result.fields[1].extra_attrs["cardinality"] == 90

    # Double stats check
    assert result.fields[2].extra_attrs["name"] == "col3"
    assert result.fields[2].extra_attrs["low"] == 0.1
    assert result.fields[2].extra_attrs["high"] == 100.1
    assert result.fields[2].extra_attrs["null_count"] == 15
    assert result.fields[2].extra_attrs["cardinality"] == 85

    # String stats check
    assert result.fields[3].extra_attrs["name"] == "col4"
    assert result.fields[3].extra_attrs["average_length"] == 10
    assert result.fields[3].extra_attrs["max_length"] == 100
    assert result.fields[3].extra_attrs["null_count"] == 20
    assert result.fields[3].extra_attrs["cardinality"] == 80

    # Binary stats check
    assert result.fields[4].extra_attrs["name"] == "col5"
    assert result.fields[4].extra_attrs["average_length"] == 500
    assert result.fields[4].extra_attrs["max_length"] == 1000
    assert result.fields[4].extra_attrs["null_count"] == 30

    # Date stats check
    assert result.fields[5].extra_attrs["name"] == "col6"
    assert result.fields[5].extra_attrs["low"] == 123
    assert result.fields[5].extra_attrs["high"] == 456
    assert result.fields[5].extra_attrs["null_count"] == 40
    assert result.fields[5].extra_attrs["cardinality"] == 90

    # Decimal stats check
    assert result.fields[6].extra_attrs["name"] == "col7"
    assert result.fields[6].extra_attrs["low"] == Decimal("0.1")
    assert result.fields[6].extra_attrs["high"] == Decimal("100.1")
    assert result.fields[6].extra_attrs["null_count"] == 50
    assert result.fields[6].extra_attrs["cardinality"] == 95


def test_get_table_stats_empty():
    mock_client = MagicMock(spec=HMS)

    class MockTable:
        name = "dummy_table"
        columns = [
            HColumn("col1", HPrimitiveType(PrimitiveCategory.BOOLEAN)),
            HColumn("col2", HPrimitiveType(PrimitiveCategory.LONG)),
            HColumn("col3", HPrimitiveType(PrimitiveCategory.DOUBLE)),
            HColumn("col4", HPrimitiveType(PrimitiveCategory.STRING)),
            HColumn("col5", HPrimitiveType(PrimitiveCategory.BINARY)),
            HColumn("col6", HPrimitiveType(PrimitiveCategory.DATE)),
        ]

    mock_client.get_table.return_value = MockTable
    mock_client.get_table_stats.return_value = []

    client = HiveMetastoreClient(mock_client)
    result = client.schema("dummy_database", "dummy_table", True)

    assert isinstance(result, StructType)
    assert len(result.fields) == 6

    for field in result.fields:
        assert len(field.extra_attrs) == 2
        assert "name" in field.extra_attrs
        assert "default" in field.extra_attrs


@pytest.mark.parametrize(
    "method, url, host, port, paths, include_stats, expected_result",
    [
        # Test ls method with only URL
        (
            "ls",
            "thrift://metastore-url",
            "metastore-url",
            None,
            [],
            None,
            ("thrift://metastore-url", []),
        ),
        # Test ls method with URL and default port
        (
            "ls",
            "thrift://metastore-url:9083",
            "metastore-url",
            9083,
            [],
            None,
            ("thrift://metastore-url:9083", []),
        ),
        # Test ls method with URL and custom port
        (
            "ls",
            "thrift://metastore-url:1234",
            "metastore-url",
            1234,
            [],
            None,
            ("thrift://metastore-url:1234", []),
        ),
        # Test ls method with URL and database
        (
            "ls",
            "thrift://metastore-url",
            "metastore-url",
            None,
            ["db1"],
            None,
            ("thrift://metastore-url", ["db1"]),
        ),
        # Test ls method with URL, database and table
        (
            "ls",
            "thrift://metastore-url",
            "metastore-url",
            None,
            ["db1", "table1"],
            None,
            ("thrift://metastore-url", ["db1", "table1"]),
        ),
        # Test schema method without include_stats
        (
            "schema",
            "thrift://metastore-url",
            "metastore-url",
            None,
            ["db1", "table1"],
            None,
            ("thrift://metastore-url", ["db1", "table1", False]),
        ),
        # Test schema method with include_stats
        (
            "schema",
            "thrift://metastore-url",
            "metastore-url",
            None,
            ["db1", "table1"],
            "some_stat",
            ("thrift://metastore-url", ["db1", "table1", True]),
        ),
        # Test invalid method
        (
            "invalid_method",
            "thrift://metastore-url",
            "metastore-url",
            None,
            [],
            None,
            pytest.raises(ValueError, match="Invalid method"),
        ),
        # Test schema method with insufficient paths
        (
            "schema",
            "thrift://metastore-url",
            "metastore-url",
            None,
            ["db1"],
            None,
            pytest.raises(ValueError, match="Invalid method"),
        ),
    ],
)
def test_parse_method(method, url, host, port, paths, include_stats, expected_result):
    if isinstance(expected_result, tuple):
        result = HiveMetastoreClient.parse(
            method,
            url,
            paths,
            include_stats,
            host=host,
            port=port,
        )
        assert result == expected_result
    else:
        with expected_result:
            HiveMetastoreClient.parse(
                method, url, paths, include_stats, host=host, port=port
            )



import json
from pathlib import Path
from urllib.parse import quote_plus, unquote_plus

import fsspec

from recap.types import RecapType, from_dict, to_dict


class RegistryStorage:
    def __init__(self, storage_url: str, **storage_url_args):
        self.fs, self.root_path = fsspec.core.url_to_fs(storage_url, **storage_url_args)
        self.fs.mkdirs(self.root_path, exist_ok=True)

    def ls(self) -> list[str]:
        return sorted(
            [
                unquote_plus(file_path[len(self.root_path) + 1 :])
                for file_path in self.fs.ls(self.root_path)
            ]
        )

    def get(
        self,
        name: str,
        version: int | None = None,
    ) -> tuple[RecapType, int] | None:
        quoted_name = quote_plus(name)

        if version is None:
            versions = self.versions(name)
            if not versions:
                return None
            version = max(versions)

        try:
            with self.fs.open(f"{self.root_path}/{quoted_name}/{version}.json") as f:
                type_json = json.load(f)
                type_ = from_dict(type_json)
                return (type_, version)
        except FileNotFoundError:
            return None

    def put(
        self,
        name: str,
        type_: RecapType,
        version: int | None = None,
    ) -> int:
        quoted_name = quote_plus(name)

        if version is None:
            version = (self.latest(name) or 0) + 1

        path_without_version = f"{self.root_path}/{quoted_name}"
        type_dict = to_dict(type_)

        self.fs.mkdirs(path_without_version, exist_ok=True)

        with self.fs.open(f"{path_without_version}/{version}.json", "w") as f:
            json.dump(type_dict, f)

        return version

    def versions(self, name: str) -> list[int] | None:
        quoted_name = quote_plus(name)
        path_without_version = f"{self.root_path}/{quoted_name}"

        try:
            return sorted(
                [
                    int(Path(file_path).stem)
                    for file_path in self.fs.ls(path_without_version)
                ]
            )
        except FileNotFoundError:
            return None

    def latest(self, name: str) -> int | None:
        versions = self.versions(name)
        if not versions:
            return None
        return max(versions)



from typing import Annotated, Optional

import typer
from rich import print, print_json

from recap import commands

app = typer.Typer()


@app.command()
def ls(url: Annotated[Optional[str], typer.Argument(help="URL to parent.")] = None):
    """
    List a URL's children.
    """

    if children := commands.ls(url, strict=False):
        print_json(data=children)


@app.command()
def schema(
    url: Annotated[str, typer.Argument(help="URL to schema.")],
    output_format: Annotated[
        commands.SchemaFormat,
        typer.Option("--output-format", "-of", help="Schema output format."),
    ] = commands.SchemaFormat.recap,
):
    """
    Get a URL's schema.
    """

    struct_obj = commands.schema(url, output_format, strict=False)
    match struct_obj:
        case dict():
            print_json(data=struct_obj)
        case str():
            print(struct_obj)
        case _:
            raise ValueError(f"Unexpected schema type: {type(struct_obj)}")


@app.command()
def serve(
    host: str = "127.0.0.1",
    port: int = 8000,
    log_level: str = "info",
):
    """
    Start Recap's HTTP/JSON gateway server.
    """

    import uvicorn

    uvicorn.run("recap.server.app:app", host=host, port=port, log_level=log_level)



from recap.clients.fs import FilesystemClient
from recap.types import IntType, StringType, StructType


def test_integration_filesystemclient_avro(tmp_path):
    # Sample Avro schema
    avro_schema_str = """
    {
      "type": "record",
      "name": "User",
      "fields": [
        {"name": "name", "type": "string"},
        {"name": "age", "type": "int"}
      ]
    }
    """

    # Create a .avsc file in the temporary directory
    avsc_file = tmp_path / "sample.avsc"
    avsc_file.write_text(avro_schema_str)

    # Create an instance of FilesystemClient with scheme "file"
    client = FilesystemClient(scheme="file")

    # Validate the schema using the client
    schema = client.schema(avsc_file.as_posix())

    # Ensure the schema is parsed and converted correctly
    assert isinstance(schema, StructType)
    assert len(schema.fields) == 2
    assert schema.fields[0].extra_attrs["name"] == "name"
    assert isinstance(schema.fields[0], StringType)
    assert schema.fields[1].extra_attrs["name"] == "age"
    assert isinstance(schema.fields[1], IntType)


def test_integration_filesystemclient_json_schema(tmp_path):
    # Sample JSON schema
    json_schema_str = """
    {
      "$schema": "http://json-schema.org/draft-07/schema#",
      "type": "object",
      "title": "User",
      "properties": {
        "name": {
          "type": "string"
        },
        "age": {
          "type": "integer"
        }
      },
      "required": ["name", "age"]
    }
    """

    # Create a .json file in the temporary directory
    json_file = tmp_path / "sample.json"
    json_file.write_text(json_schema_str)

    # Create an instance of FilesystemClient with scheme "file"
    client = FilesystemClient(scheme="file")

    # Validate the schema using the client
    schema = client.schema(json_file.as_posix())

    # Ensure the schema is parsed and converted correctly
    assert isinstance(schema, StructType)
    assert len(schema.fields) == 2
    assert schema.fields[0].extra_attrs["name"] == "name"
    assert isinstance(schema.fields[0], StringType)
    assert schema.fields[1].extra_attrs["name"] == "age"
    assert isinstance(schema.fields[1], IntType)


def test_integration_filesystemclient_proto(tmp_path):
    # Sample protobuf schema (proto3)
    proto_str = """
    syntax = "proto3";

    message User {
        required string name = 1;
        required int32 age = 2;
    }
    """

    # Create a .proto file in the temporary directory
    proto_file = tmp_path / "sample.proto"
    proto_file.write_text(proto_str)

    # Create an instance of FilesystemClient with scheme "file"
    client = FilesystemClient(scheme="file")

    # Validate the schema using the client
    schema = client.schema(proto_file.as_posix())

    # Ensure the schema is parsed and converted correctly
    assert isinstance(schema, StructType)
    assert len(schema.fields) == 2
    assert schema.fields[0].extra_attrs["name"] == "name"
    assert isinstance(schema.fields[0], StringType)
    assert schema.fields[1].extra_attrs["name"] == "age"
    assert isinstance(schema.fields[1], IntType)



import pytest
from google.cloud import bigquery

from recap.converters.bigquery import BigQueryConverter
from recap.types import BoolType, BytesType, FloatType, IntType, StringType, StructType


@pytest.mark.parametrize(
    "field_type,expected",
    [
        ("STRING", StringType(bytes_=65_536, name="test_field")),
        ("BYTES", BytesType(bytes_=65_536, name="test_field")),
        ("INT64", IntType(bits=64, name="test_field")),
        ("FLOAT", FloatType(bits=64, name="test_field")),
        ("BOOLEAN", BoolType(name="test_field")),
        (
            "TIMESTAMP",
            IntType(
                logical="build.recap.Timestamp",
                bits=64,
                unit="microsecond",
                name="test_field",
            ),
        ),
        (
            "TIME",
            IntType(
                logical="build.recap.Time",
                bits=32,
                unit="microsecond",
                name="test_field",
            ),
        ),
        (
            "DATE",
            IntType(logical="build.recap.Date", bits=32, unit="day", name="test_field"),
        ),
        (
            "NUMERIC",
            BytesType(
                logical="build.recap.Decimal",
                bytes_=16,
                variable=False,
                precision=38,
                scale=0,
                name="test_field",
            ),
        ),
        (
            "BIGNUMERIC",
            BytesType(
                logical="build.recap.Decimal",
                bytes_=32,
                variable=False,
                precision=76,
                scale=0,
                name="test_field",
            ),
        ),
    ],
)
def test_all_cases(field_type, expected):
    field = bigquery.SchemaField(
        name="test_field", field_type=field_type, mode="REQUIRED"
    )
    result = BigQueryConverter().to_recap([field])
    assert result.fields[0] == expected


def test_record():
    nested_fields = [
        bigquery.SchemaField(name="nested_int", field_type="INT64", mode="REQUIRED"),
        bigquery.SchemaField(
            name="nested_string", field_type="STRING", mode="REQUIRED"
        ),
    ]

    field = bigquery.SchemaField(
        name="test_record", field_type="RECORD", mode="REQUIRED", fields=nested_fields
    )
    expected = StructType(
        [
            IntType(bits=64, name="nested_int"),
            StringType(bytes_=65_536, name="nested_string"),
        ],
        name="test_record",
    )

    result = BigQueryConverter().to_recap([field])
    assert result.fields[0] == expected


def test_unrecognized():
    with pytest.raises(ValueError):
        field = bigquery.SchemaField(
            name="test_field", field_type="UNKNOWN", mode="REQUIRED"
        )
        BigQueryConverter().to_recap([field])



import pytest

from recap.clients.mysql import MysqlClient


@pytest.mark.parametrize(
    "method, url_args, expected_result",
    [
        # Test "ls" method with only URL (no schema)
        (
            "ls",
            {"scheme": "mysql", "netloc": "mysql-url"},
            ("mysql://mysql-url", [None]),
        ),
        # Test "ls" method with URL and schema
        (
            "ls",
            {"scheme": "mysql", "netloc": "mysql-url", "paths": ["db1"]},
            ("mysql://mysql-url/db1", ["db1"]),
        ),
        # Test "schema" method with URL, schema, and table
        (
            "schema",
            {"scheme": "mysql", "netloc": "mysql-url", "paths": ["db1", "table1"]},
            ("mysql://mysql-url/db1", ["db1", "table1"]),
        ),
        # Test invalid method
        (
            "invalid_method",
            {"scheme": "mysql", "netloc": "mysql-url"},
            pytest.raises(ValueError, match="Invalid method"),
        ),
        # Test "schema" method with insufficient paths (only schema, no table)
        (
            "schema",
            {"scheme": "mysql", "netloc": "mysql-url", "paths": ["db1"]},
            pytest.raises(ValueError, match="Invalid method"),
        ),
    ],
)
def test_parse_method(method, url_args, expected_result):
    if isinstance(expected_result, tuple):
        result = MysqlClient.parse(method, **url_args)
        assert result == expected_result
    else:
        with expected_result:
            MysqlClient.parse(method, **url_args)



import pytest
from fastapi import HTTPException
from fastapi.testclient import TestClient
from pydantic import AnyUrl

from recap.server.registry import router, settings
from recap.types import IntType, StructType, to_dict

client = TestClient(router)


@pytest.fixture(autouse=True)
def set_env_variable(tmp_path):
    settings.registry_storage_url = AnyUrl(tmp_path.as_uri())


def test_ls_empty():
    response = client.get("/registry/")
    assert response.status_code == 200
    assert response.json() == []


def test_post_and_get_latest():
    name = "test"
    type_ = StructType(fields=[IntType(bits=24)])

    # Post a new type
    response = client.post(
        f"/registry/{name}",
        json=to_dict(type_),
        headers={"Content-Type": "application/x-recap+json"},
    )
    assert response.status_code == 200
    assert response.text == "1"

    # Get the latest type
    response = client.get(f"/registry/{name}")
    assert response.status_code == 200
    data = response.json()
    assert to_dict(type_) == data[0]


def test_post_and_get_specific_version():
    name = "test_version"
    type1 = StructType(fields=[IntType(bits=32)])
    type2 = StructType(fields=[IntType(bits=24)])

    # Post two types
    response = client.post(
        f"/registry/{name}",
        json=to_dict(type1),
        headers={"Content-Type": "application/x-recap+json"},
    )
    assert response.status_code == 200
    assert response.text == "1"
    response = client.post(
        f"/registry/{name}",
        json=to_dict(type2),
        headers={"Content-Type": "application/x-recap+json"},
    )
    assert response.status_code == 200
    assert response.text == "2"

    # Get version 1
    response = client.get(f"/registry/{name}/versions/1")
    assert response.status_code == 200
    type_dict, version = response.json()
    assert to_dict(type1) == type_dict
    assert version == 1

    # Get version 2
    response = client.get(f"/registry/{name}/versions/2")
    assert response.status_code == 200
    type_dict, version = response.json()
    assert to_dict(type2) == type_dict
    assert version == 2


def test_versions_endpoint():
    name = "test_versions_endpoint"
    type1 = StructType(fields=[IntType(bits=32)])
    type2 = StructType(fields=[IntType(bits=24)])
    type3 = StructType(fields=[IntType(bits=16)])

    # Post three types for the same name
    response = client.post(
        f"/registry/{name}",
        json=to_dict(type1),
        headers={"Content-Type": "application/x-recap+json"},
    )
    assert response.status_code == 200
    assert response.text == "1"

    response = client.post(
        f"/registry/{name}",
        json=to_dict(type2),
        headers={"Content-Type": "application/x-recap+json"},
    )
    assert response.status_code == 200
    assert response.text == "2"

    response = client.post(
        f"/registry/{name}",
        json=to_dict(type3),
        headers={"Content-Type": "application/x-recap+json"},
    )
    assert response.status_code == 200
    assert response.text == "3"

    # Get versions for the name
    response = client.get(f"/registry/{name}/versions")
    assert response.status_code == 200
    versions = response.json()
    assert versions == [1, 2, 3]


def test_ls_non_empty():
    name1 = "test_schema1"
    name2 = "test_schema2"
    type1 = StructType(fields=[IntType(bits=16)])
    type2 = StructType(fields=[IntType(bits=8)])

    # Post two types under different names
    response = client.post(
        f"/registry/{name1}",
        json=to_dict(type1),
        headers={"Content-Type": "application/x-recap+json"},
    )
    assert response.status_code == 200

    response = client.post(
        f"/registry/{name2}",
        json=to_dict(type2),
        headers={"Content-Type": "application/x-recap+json"},
    )
    assert response.status_code == 200

    # Get the list of schemas
    response = client.get("/registry/")
    assert response.status_code == 200
    schema_names = response.json()
    assert schema_names == [name1, name2]


def test_put_existing_schema_conflict():
    name = "test_put_schema_conflict"
    original_type = StructType(fields=[IntType(bits=16)])
    new_type = StructType(fields=[IntType(bits=8)])

    # Post the original type
    response = client.post(
        f"/registry/{name}",
        json=to_dict(original_type),
        headers={"Content-Type": "application/x-recap+json"},
    )
    assert response.status_code == 200
    version = int(response.text)

    # Try to overwrite the original type with new_type for the same version
    with pytest.raises(HTTPException) as exc_info:
        client.put(
            f"/registry/{name}/versions/{version}",
            json=to_dict(new_type),
            headers={"Content-Type": "application/x-recap+json"},
        )

    assert exc_info.value.status_code == 409

    # Fetch the schema with the same name and version to verify it remains unchanged
    response = client.get(f"/registry/{name}/versions/{version}")
    assert response.status_code == 200
    fetched_type_dict, fetched_version = response.json()

    # Confirm the fetched type is still the original_type
    assert fetched_version == version
    assert fetched_type_dict == to_dict(original_type)



from __future__ import annotations

from contextlib import contextmanager
from importlib import import_module
from typing import Any, Generator, Protocol
from urllib.parse import parse_qs, urlparse

from recap.settings import RecapSettings
from recap.types import StructType

settings = RecapSettings()
CLIENTS = {
    None: "recap.clients.fs.FilesystemClient",
    "": "recap.clients.fs.FilesystemClient",
    "file": "recap.clients.fs.FilesystemClient",
    "bigquery": "recap.clients.bigquery.BigQueryClient",
    "http+csr": "recap.clients.confluent_registry.ConfluentRegistryClient",
    "https+csr": "recap.clients.confluent_registry.ConfluentRegistryClient",
    "mysql": "recap.clients.mysql.MysqlClient",
    "postgresql": "recap.clients.postgresql.PostgresqlClient",
    "snowflake": "recap.clients.snowflake.SnowflakeClient",
    "thrift+hms": "recap.clients.hive_metastore.HiveMetastoreClient",
}


@contextmanager
def create_client(url: str) -> Generator[Client, None, None]:
    """
    Create a client from a URL. The supported clients are listed in the
    `recap.clients.CLIENTS` dict.

    :param url: URL to create client from
    :return: Client
    """

    url_args = _url_to_dict(url)
    scheme = url_args["scheme"]

    if client_path := CLIENTS.get(scheme):
        module_path, class_name = client_path.rsplit(".", 1)
        module = import_module(module_path)
        client_class = getattr(module, class_name)
        with client_class.create(**url_args) as client:
            yield client
    else:
        raise ValueError(f"No clients available for scheme: {scheme}")


def parse_url(method: str, url: str, strict: bool = True) -> tuple[str, list[Any]]:
    """
    Parse a URL into a connection URL and a list of method arguments.

    :param method: Either "ls" or "schema".
    :param url: URL to parse
    :param strict: If True, raise an error if the URL is not configured in settings.
    :return: Tuple of connection URL and list of method arguments.
    """

    url_args = _url_to_dict(url)
    scheme = url_args["scheme"]

    if client_path := CLIENTS.get(scheme):
        module_path, class_name = client_path.rsplit(".", 1)
        module = import_module(module_path)
        client_class = getattr(module, class_name)
        connection_url, method_args = client_class.parse(method, **url_args)
        return (settings.unsafe_url(connection_url, strict), method_args)
    else:
        raise ValueError(f"No clients available for scheme: {scheme}")


def _url_to_dict(url: str) -> dict[str, str]:
    """
    Parse a URL into a dict of its components. This is a wrapper around the
    `urllib.parse.urlparse` function that adds some additional fields:

    - `dialect`: The dialect of the URL, e.g. `http`, `https`, `mysql`, etc.
    - `driver`: The driver of the URL, e.g. `csr` for Confluent Schema Registry
    - `user`: The username of the URL
    - `host`: The hostname of the URL
    - `paths`: The path of the URL split into a list
    - `url`: The original URL

    All query parameters are parsed into a dict, as well.

    :param url: URL to parse
    :return: Dict of URL components
    """

    parsed_url = urlparse(url)
    parsed_dict = {
        field: getattr(parsed_url, field)
        for field in [
            "scheme",
            "netloc",
            "path",
            "params",
            "query",
            "fragment",
            "username",
            "password",
            "hostname",
            "port",
        ]
    }

    if "+" in parsed_dict["scheme"]:
        dialect, driver = parsed_dict["scheme"].split("+")
        parsed_dict["dialect"] = dialect
        parsed_dict["driver"] = driver
        url = url.replace(f"{dialect}+{driver}://", f"{dialect}://")

    if username := parsed_url.username:
        parsed_dict["user"] = username

    if hostname := parsed_url.hostname:
        parsed_dict["host"] = hostname

    parsed_dict["path"] = parsed_dict["path"].rstrip("/")
    parsed_dict["paths"] = [path for path in parsed_dict["path"].split("/") if path]
    parsed_dict["url"] = url
    parsed_qs = parse_qs(parsed_dict["query"])
    parsed_qs = {
        key: value[0] if len(value) == 1 else value for key, value in parsed_qs.items()
    }
    parsed_dict |= parsed_qs

    return parsed_dict


class Client(Protocol):
    @staticmethod
    def create(**url_args) -> Client:
        ...

    @staticmethod
    def parse(method: str, **url_args) -> tuple[str, list[Any]]:
        ...

    def ls(self, *vargs) -> list[str]:
        ...

    def schema(self, *vargs) -> StructType:
        ...



from __future__ import annotations

from contextlib import contextmanager
from typing import Generator

from recap.clients.dbapi import Connection, DbapiClient
from recap.converters.snowflake import SnowflakeConverter


class SnowflakeClient(DbapiClient):
    def __init__(self, connection: Connection) -> None:
        super().__init__(connection, SnowflakeConverter())

    @staticmethod
    @contextmanager
    def create(**url_args) -> Generator[SnowflakeClient, None, None]:
        import snowflake.connector

        snowflake_args = {}
        snowflake_args["account"] = url_args.pop("host")

        match url_args.get("paths"):
            case str(database), None:
                snowflake_args["database"] = database
            case str(database), str(schema):
                snowflake_args["database"] = database
                snowflake_args["schema"] = schema

        with snowflake.connector.connect(**(snowflake_args | url_args)) as client:
            yield SnowflakeClient(client)  # type: ignore

    def ls_catalogs(self) -> list[str]:
        cursor = self.connection.cursor()
        cursor.execute(
            """
                SELECT database_name
                FROM information_schema.databases
                ORDER BY database_name ASC
            """
        )
        return [row[0] for row in cursor.fetchall()]



# pylint: disable=missing-docstring

import json

import pytest

from recap.converters.avro import AvroConverter
from recap.types import (
    BoolType,
    BytesType,
    EnumType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    ProxyType,
    StringType,
    StructType,
    UnionType,
    from_dict,
)


def test_to_recap_parse_primitives():
    converter = AvroConverter()
    primitives = [
        ("null", NullType()),
        ("boolean", BoolType()),
        ("int", IntType(32, signed=True)),
        ("long", IntType(64, signed=True)),
        ("float", FloatType(32)),
        ("double", FloatType(64)),
        ("bytes", BytesType(9_223_372_036_854_775_807, variable=True)),
        ("string", StringType(9_223_372_036_854_775_807, variable=True)),
    ]
    for avro_type, recap_type in primitives:
        assert converter._parse(avro_type, "_root") == recap_type


def test_to_recap_fixed():
    converter = AvroConverter()
    avro_enum = {
        "type": "record",
        "name": "TestEnum",
        "fields": [
            {
                "name": "fixed_field",
                "type": {
                    "type": "fixed",
                    "name": "MD5",
                    "size": 16,
                },
            }
        ],
    }
    actual = converter.to_recap(json.dumps(avro_enum))

    fixed_type = BytesType(
        bytes_=16,
        name="fixed_field",
        alias="_root.MD5",
    )
    expected = StructType(fields=[fixed_type], name="TestEnum", alias="_root.TestEnum")

    assert actual == expected


def test_to_recap_enum():
    converter = AvroConverter()
    avro_enum = {
        "type": "record",
        "name": "TestEnum",
        "fields": [
            {
                "name": "color",
                "type": {
                    "type": "enum",
                    "name": "Color",
                    "symbols": ["RED", "GREEN", "BLUE"],
                },
            }
        ],
    }
    actual = converter.to_recap(json.dumps(avro_enum))

    enum_type = EnumType(
        symbols=["RED", "GREEN", "BLUE"],
        name="color",
        alias="_root.Color",
    )
    expected = StructType(fields=[enum_type], name="TestEnum", alias="_root.TestEnum")

    assert actual == expected


def test_to_recap_enum_with_inherited_namespace():
    converter = AvroConverter()
    avro_enum = {
        "type": "record",
        "namespace": "com.example",
        "name": "TestEnum",
        "fields": [
            {
                "name": "color",
                "type": {
                    "type": "enum",
                    "name": "Color",
                    "symbols": ["RED", "GREEN", "BLUE"],
                },
            }
        ],
    }
    actual = converter.to_recap(json.dumps(avro_enum))

    enum_type = EnumType(
        symbols=["RED", "GREEN", "BLUE"],
        name="color",
        alias="com.example.Color",
    )
    expected = StructType(
        fields=[enum_type],
        name="TestEnum",
        alias="com.example.TestEnum",
    )

    assert actual == expected


def test_to_recap_enum_with_namespace_override():
    converter = AvroConverter()
    avro_enum = {
        "type": "record",
        "namespace": "com.example",
        "name": "TestEnum",
        "fields": [
            {
                "name": "color",
                "type": {
                    "type": "enum",
                    "namespace": "foo.bar",
                    "name": "Color",
                    "symbols": ["RED", "GREEN", "BLUE"],
                },
            }
        ],
    }
    actual = converter.to_recap(json.dumps(avro_enum))

    enum_type = EnumType(
        symbols=["RED", "GREEN", "BLUE"],
        name="color",
        alias="foo.bar.Color",
    )
    expected = StructType(
        fields=[enum_type],
        name="TestEnum",
        alias="com.example.TestEnum",
    )

    assert actual == expected


def test_to_recap_enum_with_fully_qualified_name():
    converter = AvroConverter()
    avro_enum = {
        "type": "record",
        "name": "com.example.TestEnum",
        "fields": [
            {
                "name": "color",
                "type": {
                    "type": "enum",
                    "name": "Color",
                    "symbols": ["RED", "GREEN", "BLUE"],
                },
            }
        ],
    }
    actual = converter.to_recap(json.dumps(avro_enum))

    enum_type = EnumType(
        symbols=["RED", "GREEN", "BLUE"],
        name="color",
        alias="com.example.Color",
    )
    expected = StructType(
        fields=[enum_type],
        name="TestEnum",
        alias="com.example.TestEnum",
    )

    assert actual == expected


def test_to_recap_enum_with_fully_qualified_name_and_override():
    converter = AvroConverter()
    avro_enum = {
        "type": "record",
        "name": "com.example.TestEnum",
        "fields": [
            {
                "name": "color",
                "type": {
                    "type": "enum",
                    "name": "foo.bar.Color",
                    "symbols": ["RED", "GREEN", "BLUE"],
                },
            }
        ],
    }
    actual = converter.to_recap(json.dumps(avro_enum))

    enum_type = EnumType(
        symbols=["RED", "GREEN", "BLUE"],
        name="color",
        alias="foo.bar.Color",
    )
    expected = StructType(
        fields=[enum_type],
        name="TestEnum",
        alias="com.example.TestEnum",
    )

    assert actual == expected


def test_to_recap_array():
    converter = AvroConverter()
    avro_array = {
        "type": "record",
        "name": "TestArray",
        "fields": [{"name": "items", "type": {"type": "array", "items": "string"}}],
    }
    actual = converter.to_recap(json.dumps(avro_array))

    array_type = ListType(name="items", values=StringType(9_223_372_036_854_775_807))
    expected = StructType(
        fields=[array_type], name="TestArray", alias="_root.TestArray"
    )

    assert actual == expected


def test_to_recap_map():
    converter = AvroConverter()
    avro_map = {
        "type": "record",
        "name": "TestMap",
        "fields": [{"name": "values", "type": {"type": "map", "values": "int"}}],
    }
    actual = converter.to_recap(json.dumps(avro_map))

    # Construct an expected Map RecapType instance
    map_type = MapType(
        name="values",
        keys=StringType(9_223_372_036_854_775_807),
        values=IntType(32, signed=True),
    )
    expected = StructType(fields=[map_type], name="TestMap", alias="_root.TestMap")

    assert actual == expected


def test_to_recap_union():
    converter = AvroConverter()
    avro_union = {
        "type": "record",
        "name": "TestUnion",
        "fields": [
            {
                "name": "unionField",
                "type": {"type": "union", "types": ["string", "null"]},
            }
        ],
    }
    actual = converter.to_recap(json.dumps(avro_union))

    # Construct an expected Union RecapType instance
    union_type = UnionType(
        name="unionField",
        types=[StringType(9_223_372_036_854_775_807), NullType()],
    )
    expected = StructType(
        fields=[union_type],
        name="TestUnion",
        alias="_root.TestUnion",
    )

    assert actual == expected


def test_to_recap_record():
    converter = AvroConverter()
    avro_record = {
        "type": "record",
        "name": "Test",
        "fields": [
            {"name": "a", "type": "int"},
            {"name": "b", "type": "string"},
        ],
    }
    actual = converter.to_recap(json.dumps(avro_record))
    assert isinstance(actual, StructType)
    assert len(actual.fields) == 2
    assert isinstance(actual.fields[0], IntType)
    assert actual.fields[0].bits == 32
    assert actual.fields[0].extra_attrs["name"] == "a"
    assert isinstance(actual.fields[1], StringType)
    assert actual.fields[1].bytes_ == 9_223_372_036_854_775_807
    assert actual.fields[1].extra_attrs["name"] == "b"
    assert actual.extra_attrs == {"name": "Test"}
    assert actual.alias == "_root.Test"


def test_to_recap_self_referencing():
    converter = AvroConverter()

    # Avro schema for a simple linked list
    avro_schema = {
        "type": "record",
        "name": "build.recap.LinkedList",
        "fields": [
            {"name": "value", "type": "int"},
            {
                "name": "next",
                "type": ["null", "build.recap.LinkedList"],
            },
        ],
    }

    # The expected schema is a recursive structure, so we can't create it
    # directly as Python would run into an infinite recursion. We can only
    # check that the resulting structure has the expected properties.
    actual = converter.to_recap(json.dumps(avro_schema))

    assert isinstance(actual, StructType)
    assert actual.extra_attrs["name"] == "LinkedList"
    assert actual.alias == "build.recap.LinkedList"
    assert len(actual.fields) == 2
    assert isinstance(actual.fields[0], IntType)
    assert isinstance(actual.fields[1], UnionType)
    assert actual.fields[1].extra_attrs["name"] == "next"
    assert isinstance(actual.fields[1].types[0], NullType)
    assert isinstance(actual.fields[1].types[1], ProxyType)
    assert actual.fields[1].types[1].alias == "build.recap.LinkedList"

    # Strip the alias since the resolved type won't have an alias (since it's
    # already been defined in the avro_schema struct).
    actual.alias = None

    assert actual.fields[1].types[1].resolve() == actual


def test_to_recap_record_with_docs_and_default():
    converter = AvroConverter()

    # Avro schema for a record with docs and default attributes
    avro_schema = {
        "type": "record",
        "name": "Person",
        "doc": "A person record",
        "fields": [
            {
                "name": "name",
                "type": "string",
                "doc": "The person's name",
                "default": "unknown",
            },
            {"name": "age", "type": "int", "doc": "The person's age", "default": 0},
        ],
    }

    actual = converter.to_recap(json.dumps(avro_schema))

    assert isinstance(actual, StructType)
    assert len(actual.fields) == 2
    assert actual.doc == "A person record"
    assert isinstance(actual.fields[0], StringType)
    assert actual.fields[0].doc == "The person's name"
    assert actual.fields[0].extra_attrs["default"] == "unknown"
    assert isinstance(actual.fields[1], IntType)
    assert actual.fields[1].doc == "The person's age"
    assert actual.fields[1].extra_attrs["default"] == 0


def test_to_recap_decimal():
    converter = AvroConverter()
    avro_schema = {
        "type": "record",
        "name": "test_decimal",
        "fields": [
            {
                "name": "decimal",
                "type": {
                    "type": "bytes",
                    "logicalType": "decimal",
                    "precision": 5,
                    "scale": 2,
                },
            }
        ],
    }
    schema = converter.to_recap(json.dumps(avro_schema))
    field = schema.fields[0]
    assert isinstance(field, BytesType)
    assert field.logical == "build.recap.Decimal"
    assert field.bytes_ == avro_schema["fields"][0]["type"].get(
        "size",
        9_223_372_036_854_775_807,
    )
    assert field.variable == (avro_schema["fields"][0]["type"]["type"] == "fixed")
    assert (
        field.extra_attrs["precision"] == avro_schema["fields"][0]["type"]["precision"]
    )
    assert field.extra_attrs["scale"] == avro_schema["fields"][0]["type"].get(
        "scale",
        0,
    )


def test_to_recap_uuid():
    converter = AvroConverter()
    avro_schema = {
        "type": "record",
        "name": "test_uuid",
        "fields": [{"name": "uuid", "type": {"type": "string", "logicalType": "uuid"}}],
    }
    schema = converter.to_recap(json.dumps(avro_schema))
    field = schema.fields[0]
    assert isinstance(field, StringType)
    assert field.logical == "build.recap.UUID"
    assert field.bytes_ == 36
    assert field.variable == False


def test_to_recap_date():
    converter = AvroConverter()
    avro_schema = {
        "type": "record",
        "name": "test_date",
        "fields": [{"name": "date", "type": {"type": "int", "logicalType": "date"}}],
    }
    schema = converter.to_recap(json.dumps(avro_schema))
    field = schema.fields[0]
    assert isinstance(field, IntType)
    assert field.logical == "build.recap.Date"
    assert field.bits == 32
    assert field.signed is True
    assert field.extra_attrs["unit"] == "day"


def test_to_recap_time_millis():
    converter = AvroConverter()
    avro_schema = {
        "type": "record",
        "name": "test_time_millis",
        "fields": [
            {"name": "time", "type": {"type": "int", "logicalType": "time-millis"}}
        ],
    }
    schema = converter.to_recap(json.dumps(avro_schema))
    field = schema.fields[0]
    assert isinstance(field, IntType)
    assert field.logical == "build.recap.Time"
    assert field.bits == 32
    assert field.signed is True
    assert field.extra_attrs["unit"] == "millisecond"


def test_to_recap_time_micros():
    converter = AvroConverter()
    avro_schema = {
        "type": "record",
        "name": "test_time_micros",
        "fields": [
            {"name": "time", "type": {"type": "long", "logicalType": "time-micros"}}
        ],
    }
    schema = converter.to_recap(json.dumps(avro_schema))
    field = schema.fields[0]
    assert isinstance(field, IntType)
    assert field.logical == "build.recap.Time"
    assert field.bits == 64
    assert field.signed is True
    assert field.extra_attrs["unit"] == "microsecond"


def test_to_recap_timestamp_millis():
    converter = AvroConverter()
    avro_schema = {
        "type": "record",
        "name": "test_timestamp_millis",
        "fields": [
            {
                "name": "timestamp",
                "type": {"type": "long", "logicalType": "timestamp-millis"},
            }
        ],
    }
    schema = converter.to_recap(json.dumps(avro_schema))
    field = schema.fields[0]
    assert isinstance(field, IntType)
    assert field.logical == "build.recap.Timestamp"
    assert field.bits == 64
    assert field.signed is True
    assert field.extra_attrs["unit"] == "millisecond"


def test_to_recap_timestamp_micros():
    converter = AvroConverter()
    avro_schema = {
        "type": "record",
        "name": "test_timestamp_micros",
        "fields": [
            {
                "name": "timestamp",
                "type": {"type": "long", "logicalType": "timestamp-micros"},
            }
        ],
    }
    schema = converter.to_recap(json.dumps(avro_schema))
    field = schema.fields[0]
    assert isinstance(field, IntType)
    assert field.logical == "build.recap.Timestamp"
    assert field.bits == 64
    assert field.signed is True
    assert field.extra_attrs["unit"] == "microsecond"


def test_to_recap_duration():
    converter = AvroConverter()
    avro_schema = {
        "type": "record",
        "name": "test_duration",
        "fields": [
            {
                "name": "duration",
                "type": {"type": "fixed", "size": 12, "logicalType": "duration"},
            }
        ],
    }
    schema = converter.to_recap(json.dumps(avro_schema))
    field = schema.fields[0]
    assert isinstance(field, BytesType)
    assert field.logical == "build.recap.Interval"
    assert field.bytes_ == 12
    assert field.extra_attrs["unit"] == "millisecond"


@pytest.mark.parametrize(
    "avro_type_dict, recap_type",
    [
        ({"type": "null"}, NullType()),
        ({"type": "null"}, NullType(some_attr="some_value")),
        ({"type": "boolean"}, BoolType()),
        ({"type": "int"}, IntType(bits=32, signed=True)),
        ({"type": "long"}, IntType(bits=64, signed=True)),
        ({"type": "float"}, FloatType(bits=32)),
        ({"type": "double"}, FloatType(bits=64)),
        (
            {"type": "bytes"},
            BytesType(
                bytes_=9_223_372_036_854_775_807,
                variable=True,
            ),
        ),
        (
            {"type": "string"},
            StringType(
                bytes_=9_223_372_036_854_775_807,
                variable=True,
            ),
        ),
        (
            {"type": "map", "values": "int"},
            MapType(
                keys=StringType(bytes_=9_223_372_036_854_775_807, variable=True),
                values=IntType(bits=32, signed=True),
            ),
        ),
        (
            {"type": "array", "items": "long"},
            ListType(values=IntType(bits=64, signed=True)),
        ),
        (
            {
                "type": "record",
                "fields": [{"type": "int", "name": "field1"}],
            },
            StructType(fields=[IntType(bits=32, signed=True, name="field1")]),
        ),
        (
            {"type": "enum", "symbols": ["A", "B", "C"]},
            EnumType(symbols=["A", "B", "C"]),
        ),
        (
            {"type": ["null", "int"]},
            UnionType(types=[NullType(), IntType(bits=32, signed=True)]),
        ),
        (
            {"type": "long"},
            IntType(bits=32, signed=False),
        ),
        (
            {"type": "fixed", "size": 4, "name": "fixed"},
            BytesType(bytes_=4, variable=False, name="fixed"),
        ),
        (
            {
                "type": "record",
                "fields": [
                    {
                        "name": "field1",
                        "type": "array",
                        "items": {
                            "type": "map",
                            "values": "int",
                        },
                    },
                ],
            },
            StructType(
                fields=[
                    ListType(
                        name="field1",
                        values=MapType(
                            keys=StringType(),
                            values=IntType(bits=32, signed=True),
                        ),
                    )
                ],
            ),
        ),
        (
            {
                "type": "bytes",
                "logicalType": "decimal",
                "precision": 38,
            },
            IntType(bits=128, signed=True),
        ),
        (
            {
                "type": "bytes",
                "logicalType": "decimal",
                "precision": 38,
                "scale": 0,
            },
            BytesType(
                logical="build.recap.Decimal",
                bytes_=9_223_372_036_854_775_807,
                variable=True,
                precision=38,
                scale=0,
            ),
        ),
        (
            {
                "type": "string",
                "logicalType": "uuid",
            },
            StringType(
                logical="build.recap.UUID",
                bytes_=36,
                variable=False,
            ),
        ),
        (
            {
                "type": "int",
                "logicalType": "date",
            },
            IntType(
                logical="build.recap.Date",
                bits=32,
                signed=True,
                unit="day",
            ),
        ),
        (
            {
                "type": "int",
                "logicalType": "time-millis",
            },
            IntType(
                logical="build.recap.Time",
                bits=32,
                signed=True,
                unit="millisecond",
            ),
        ),
        (
            {
                "type": "long",
                "logicalType": "time-micros",
            },
            IntType(
                logical="build.recap.Time",
                bits=64,
                signed=True,
                unit="microsecond",
            ),
        ),
        (
            {
                "type": "long",
                "logicalType": "timestamp-millis",
            },
            IntType(
                logical="build.recap.Timestamp",
                bits=64,
                signed=True,
                unit="millisecond",
            ),
        ),
        (
            {
                "type": "long",
                "logicalType": "timestamp-micros",
            },
            IntType(
                logical="build.recap.Timestamp",
                bits=64,
                signed=True,
                unit="microsecond",
            ),
        ),
        (
            # Not valid Avro, since Fixed requires a name. This is fine since
            # we are assuming fixed, enum, etc are in a StructType with a name.
            {
                "type": "fixed",
                "size": 12,
                "logicalType": "duration",
            },
            BytesType(
                logical="build.recap.Interval",
                bytes_=12,
                variable=False,
                unit="millisecond",
            ),
        ),
        (
            # Test a logical type with extra attributes.
            {
                "type": "long",
                "logicalType": "timestamp-micros",
            },
            IntType(
                logical="build.recap.Timestamp",
                bits=64,
                signed=True,
                unit="microsecond",
                some_attr="foo",
            ),
        ),
        (
            {
                "type": "record",
                "aliases": ["build.recap.TestAlias"],
                "fields": [{"type": "int", "name": "field1"}],
            },
            StructType(
                alias="build.recap.TestAlias",
                fields=[
                    IntType(bits=32, signed=True, name="field1"),
                ],
            ),
        ),
    ],
)
def test_from_recap_primitives(avro_type_dict, recap_type):
    assert AvroConverter().from_recap(recap_type) == avro_type_dict


def test_from_recap_proxy_types_builtin_int():
    recap_type = from_dict(
        {
            "type": "struct",
            "fields": [
                {
                    "name": "field1",
                    "type": "int32",
                },
            ],
        }
    )
    avro_schema = AvroConverter().from_recap(recap_type)
    assert avro_schema == {
        "type": "record",
        "fields": [
            {
                "name": "field1",
                "type": "int",
            },
        ],
    }


def test_from_recap_proxy_types_regular():
    recap_type = from_dict(
        {
            "type": "struct",
            "fields": [
                {
                    "name": "field1",
                    "alias": "build.recap.Field1Alias",
                    "type": "int",
                    "bits": 32,
                },
                {
                    "name": "field2",
                    "type": "build.recap.Field1Alias",
                },
            ],
        }
    )
    avro_schema = AvroConverter().from_recap(recap_type)
    assert avro_schema == {
        "type": "record",
        "fields": [
            {
                "name": "field1",
                "aliases": ["build.recap.Field1Alias"],
                "type": "int",
            },
            {
                "name": "field2",
                "type": "build.recap.Field1Alias",
            },
        ],
    }


def test_from_recap_proxy_types_self_reference():
    recap_type = from_dict(
        {
            "type": "struct",
            "alias": "build.recap.LinkedListNode",
            "fields": [
                {
                    "name": "value",
                    "type": "int32",
                },
                {
                    "name": "next",
                    "type": "build.recap.LinkedListNode",
                },
            ],
        }
    )
    avro_schema = AvroConverter().from_recap(recap_type)
    assert avro_schema == {
        "type": "record",
        "aliases": ["build.recap.LinkedListNode"],
        "fields": [
            {
                "name": "value",
                "type": "int",
            },
            {
                "name": "next",
                "type": "build.recap.LinkedListNode",
            },
        ],
    }



from enum import Enum

from recap.clients import create_client, parse_url
from recap.settings import RecapSettings

settings = RecapSettings()


class SchemaFormat(str, Enum):
    """
    Schema formats Recap can convert to. Used in the `schema` method.
    """

    avro = "avro"
    json = "json"
    protobuf = "protobuf"
    recap = "recap"


FORMAT_MAP = {
    "application/schema+json": SchemaFormat.json,
    "application/avro+json": SchemaFormat.avro,
    "application/x-protobuf": SchemaFormat.protobuf,
    "application/x-recap+json": SchemaFormat.recap,
}


def ls(url: str | None = None, strict: bool = True) -> list[str] | None:
    """
    List a URL's children.

    :param url: URL where children are located. If `url` is None, list root URLs.
    :param strict: If True, raise an error if the URL is not configured in settings.
    :return: List of children. Values are relative to `url`.
    """

    if not url:
        return settings.safe_urls
    connection_url, method_args = parse_url("ls", url, strict)
    with create_client(connection_url) as client:
        return client.ls(*method_args)


def schema(
    url: str,
    format: SchemaFormat = SchemaFormat.recap,
    strict: bool = True,
) -> dict | str:
    """
    Get a URL's schema.

    :param url: URL where schema is located.
    :param format: Schema format to convert to.
    :param strict: If True, raise an error if the URL is not configured in settings.
    :return: Schema in the requested format (encoded as a dict or string).
    """

    connection_url, method_args = parse_url("schema", url, strict)
    with create_client(connection_url) as client:
        recap_struct = client.schema(*method_args)
        output_obj: dict | str
        match format:
            case SchemaFormat.avro:
                from recap.converters.avro import AvroConverter

                output_obj = AvroConverter().from_recap(recap_struct)
            case SchemaFormat.json:
                from recap.converters.json_schema import JSONSchemaConverter

                output_obj = JSONSchemaConverter().from_recap(recap_struct)
            case SchemaFormat.protobuf:
                from proto_schema_parser.generator import Generator

                from recap.converters.protobuf import ProtobufConverter

                proto_file = ProtobufConverter().from_recap(recap_struct)
                proto_str = Generator().generate(proto_file)

                output_obj = proto_str
            case SchemaFormat.recap:
                from recap.types import to_dict

                struct_dict = to_dict(recap_struct)
                if not isinstance(struct_dict, dict):
                    raise ValueError(
                        f"Expected a schema dict, but got {type(struct_dict)}"
                    )
                output_obj = struct_dict
            case _:
                raise ValueError(f"Unknown schema format: {format}")
        return output_obj



from json import loads

import psycopg2
import pytest
from typer.testing import CliRunner

from recap.cli import app

runner = CliRunner()


class TestCli:
    @classmethod
    def setup_class(cls):
        # Connect to the PostgreSQL database
        cls.connection = psycopg2.connect(
            host="localhost",
            port="5432",
            user="postgres",
            password="password",
            dbname="testdb",
        )

        # Create tables
        cursor = cls.connection.cursor()
        cursor.execute("CREATE TABLE IF NOT EXISTS test_types (test_integer INTEGER);")
        cls.connection.commit()

    @classmethod
    def teardown_class(cls):
        # Delete tables
        cursor = cls.connection.cursor()
        cursor.execute("DROP TABLE IF EXISTS test_types;")
        cls.connection.commit()

        # Close the connection
        cls.connection.close()

    @pytest.mark.parametrize(
        "cmd, url, expected",
        [
            ["ls", "", ["postgresql://localhost:5432/testdb"]],
            [
                "ls",
                "postgresql://postgres:password@localhost:5432",
                ["postgres", "template0", "template1", "testdb"],
            ],
            [
                "ls",
                "postgresql://postgres:password@localhost:5432/testdb",
                [
                    "pg_toast",
                    "pg_catalog",
                    "public",
                    "information_schema",
                ],
            ],
            [
                "ls",
                "postgresql://postgres:password@localhost:5432/testdb/",
                [
                    "pg_toast",
                    "pg_catalog",
                    "public",
                    "information_schema",
                ],
            ],
            ["ls", "postgresql://localhost:5432/testdb/public", ["test_types"]],
        ],
    )
    def test_ls(self, cmd, url, expected):
        result = runner.invoke(app, [cmd, url])
        assert result.exit_code == 0
        assert loads(result.stdout) == expected

    def test_schema(self):
        result = runner.invoke(
            app,
            [
                "schema",
                "postgresql://localhost:5432/testdb/public/test_types",
            ],
        )
        assert result.exit_code == 0
        assert loads(result.stdout) == {
            "type": "struct",
            "fields": [{"type": "int32", "name": "test_integer", "optional": True}],
        }

    def test_schema_avro(self):
        result = runner.invoke(
            app,
            [
                "schema",
                "postgresql://localhost:5432/testdb/public/test_types",
                "-of=avro",
            ],
        )
        assert result.exit_code == 0
        assert loads(result.stdout) == {
            "type": "record",
            "fields": [
                {"name": "test_integer", "default": None, "type": ["null", "int"]}
            ],
        }

    def test_schema_json(self):
        result = runner.invoke(
            app,
            [
                "schema",
                "postgresql://localhost:5432/testdb/public/test_types",
                "-of=json",
            ],
        )
        assert result.exit_code == 0
        assert loads(result.stdout) == {
            "type": "object",
            "properties": {"test_integer": {"default": None, "type": "integer"}},
        }

    @pytest.mark.skip(reason="Enable when #397 is fixed")
    def test_schema_protobuf(self):
        result = runner.invoke(
            app,
            [
                "schema",
                "postgresql://localhost:5432/testdb/public/test_types",
                "-of=protobuf",
            ],
        )
        assert result.exit_code == 0
        assert (
            result.stdout
            == """
TODO: Some proto schema
"""
        )

    def test_schema_recap(self):
        result = runner.invoke(
            app,
            [
                "schema",
                "postgresql://localhost:5432/testdb/public/test_types",
                "-of=recap",
            ],
        )
        assert result.exit_code == 0
        assert loads(result.stdout) == {
            "type": "struct",
            "fields": [{"type": "int32", "name": "test_integer", "optional": True}],
        }
